/*
 * Copyright (c) 1996, 1998, 1999 University of Utah and the Flux Group.
 * All rights reserved.
 * @OSKIT-FLUX-GPLUS@
 */

#include <sys/param.h>
#include <sys/time.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "fsck.h"

#include <oskit/fs/netbsd.h>
#include <oskit/fs/filesystem.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/linux.h>
#include <oskit/io/blkio.h>
#include <oskit/fs/dir.h>
#include <oskit/diskpart/diskpart.h>
#include <oskit/diskpart/disklabel.h>
#include <oskit/startup.h>
#include <oskit/c/fd.h>
#include <oskit/c/fs.h>
#include <oskit/principal.h>
#include <oskit/boolean.h>
#include <oskit/clientos.h>
#include <oskit/startup.h>

#define MAX_PARTS 30
diskpart_t part_array[MAX_PARTS];

oskit_principal_t *cur_principal; /* identity of current client process */

/*
 * Get the world ready for fsck. The fstab argument is a boolean indicating
 * whether a specific set of filesystems is being checked, or if the fstab
 * is needed to find the filesystems.
 */
void
oskit_fsck_main(int fstab)
{
	oskit_clientos_init();
	start_clock();
	start_blk_devices();
	start_fs_bmod();
	
	if (fstab)
		dorootstuff();
	
	oskit_create_slashdev();
}

/*
 * Ask for the name of the root partition so we can mount it. It gets
 * mounted under /mnt, where we can then copy out /mnt/etc/fstab to
 * /etc/fstab (which is in the bmod filesystem). Then we unmount /mnt,
 * thereby avoind the entire issue of the hotroot. If the root partition
 * is trashed so badly that we cannot get /etc/fstab, we are going to
 * bail, and force the use to boot up BSD to get it fixed up.
 */
int
dorootstuff()
{
	char		*bp, buf[BUFSIZ];
	char		disk[8], part[8];
	oskit_blkio_t	*bio;
	oskit_dir_t	*root;
	oskit_filesystem_t *fs;
	oskit_error_t	rc;
	oskit_identity_t id;
	int		fdin, fdout, c;
	
	/*
	 * Prompt for the name of the root partition.
	 */
 again:
	printf("What is the name of the root partition"
	       " (eg, \"sd0a\", \"wd0a\", or \"quit\"): ");
		
	if ((bp = fgets(buf, BUFSIZ, stdin)) == NULL)
		panic("Problem mouting root partition\n");

	if (*buf == '\0' || *buf == '\n')
		goto again;

	if (strncmp(buf, "quit", 4) == 0)
		exit(1);

        memset(disk, 0, sizeof(disk));
        memset(part, 0, sizeof(part));
	strncpy(disk, bp, 3);
	strncpy(part, bp+3, sizeof(part));

	/*
	 * Fire it off.
	 */
	if ((rc = start_disk(disk, part, 1, &bio)) != NULL) {
		printf("dorootstuff: "
		       "start_disk() failed:  ernno 0x%x\n", rc);
		exit(rc);
	}

	/* initialize the file system code */
	if ((rc = fs_netbsd_init(start_osenv())) != NULL) {
		printf("dorootstuff: "
		       "fs_netbsd_init() failed:  errno 0x%x\n", rc);
		exit(rc);
	}

	/* establish client identity */
	id.uid = 0;
	id.gid = 0;
	id.ngroups = 0;
	id.groups = 0;
	if ((rc = oskit_principal_create(&id, &cur_principal)) != NULL) {
		printf("dorootstuff: "
		       "oskit_principal_create() failed:  errno 0x%x\n", rc);
		exit(rc);
	}

	/* mount a filesystem */
	if ((rc = fs_netbsd_mount(bio, OSKIT_FS_RDONLY, &fs)) != NULL) {
		printf("dorootstuff: "
		       "fs_netbsd_mount() failed:  errno 0x%x\n", rc);
		exit(rc);
	}

	if ((rc = oskit_filesystem_getroot(fs, &root)) != NULL) {
		printf("dorootstuff: "
		       "oskit_filesystem_getroot() failed: errno 0x%x\n", rc);
		exit(rc);
	}
	
	if (mkdir("/mnt", 0666) < 0)
		if (errno != OSKIT_EEXIST) {
			printf("dorootstuff: "
			       "Could not create /mnt: errno 0x%x\n", errno);
			exit(errno);
		}

	if ((rc = fs_mount("/mnt", root)) != NULL) {
		printf("dorootstuff: "
		       "fs_mount() failed: errno 0x%x\n", rc);
		exit(rc);
	}

	/*
	 * Okay, the root partition is mounted. Copy /mnt/etc/fstab
	 * to /etc/fstab.
	 */
	if (mkdir("/etc", 0666) < 0) 
		if (errno != OSKIT_EEXIST) {
			printf("dorootstuff: "
			       "Could not create /etc: errno 0x%x\n", errno);
			exit(errno);
		}

	if ((fdin = open("/mnt/etc/fstab", O_RDONLY, 0666)) <= 0) {
			printf("dorootstuff: "
			       "Could not open /mnt/etc/fstab: "
			       "errno 0x%x\n", errno);
			exit(errno);
	}

	if ((fdout = open("/etc/fstab", O_WRONLY|O_CREAT, 0666)) <= 0) {
			printf("dorootstuff: "
			       "Could not open /etc/fstab: "
			       "errno 0x%x\n", errno);
			exit(errno);
	}

	while (1) {
		if ((c = read(fdin, buf, BUFSIZ)) <= 0) {
			if (c == 0)
				break;
			printf("dorootstuff: "
			       "Error reading /mnt/etc/fstab: "
			       "errno 0x%x\n", errno);
			exit(errno);
		}
		if ((c = write(fdout, buf, BUFSIZ)) <= 0) {
			printf("dorootstuff: "
			       "Error writing /etc/fstab: "
			       "errno 0x%x\n", errno);
			exit(errno);
		}
	}
	close(fdin);
	close(fdout);

	if ((rc = fs_unmount("/mnt")) != NULL) {
		printf("dorootstuff: "
		       "fs_unmount() failed: errno 0x%x\n", rc);
		exit(rc);
	}
	oskit_dir_release(root);
	oskit_filesystem_release(fs);
	oskit_blkio_release(bio);

	return 0;
}

oskit_error_t
oskit_get_call_context(const struct oskit_guid *iid, void **out_if)
{
    if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
        memcmp(iid, &oskit_principal_iid, sizeof(*iid)) == 0) {
        *out_if = cur_principal;
        oskit_principal_addref(cur_principal);
        return 0;
    }

    *out_if = 0;
    return OSKIT_E_NOINTERFACE;
}

char *
blockcheck(name)
	char *name;
{
	return name;
}

/*
 * Do not need the disklabel, unless the superblock is trashed, in which case
 * better boot a real OS to deal with it.
 */
struct disklabel *
getdisklabel(s, fd)
	char *s;
	int	fd;
{
	if (s)
		panic("getdisklabel: Cannot get the disklabel for %s\n", s);
	
	return (struct disklabel *) 0;
}

void
sync()
{
}

int
fork(void)
{
	panic("fork");
}

int
wait(int *status)
{
	panic("wait");
}

unsigned int
sleep(unsigned int seconds)
{
	panic("sleep");
}
